import scipy
from scipy import ndimage
import numpy as np


def get_rotation_matrix(rotation_axis, theta, rank=3):
    r"""
    .. math::

        \begin{pmatrix}
            \cos(\frac{x'}{x})&\cos(\frac{y'}{x})&\cos(\frac{z'}{x})\\
            \cos(\frac{x'}{y})&\cos(\frac{y'}{y})&\cos(\frac{z'}{y})\\
            \cos(\frac{x'}{z})&\cos(\frac{y'}{z})&\cos(\frac{z'}{z})
        \end{pmatrix}

    x, y, z are axis 2, 0, 1 respectively

    Parameters
    ----------
    rotation_axis : [type]
        [description]
    theta : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    theta = np.deg2rad(theta)
    if rank == 3:
        if rotation_axis == 0:  # z
            rotation_matrix = np.array([[1, 0, 0],
                                        [0, np.cos(theta), -np.sin(theta)],
                                        [0, np.sin(theta), np.cos(theta)]])

        elif rotation_axis == 1:  # x
            rotation_matrix = np.array([[np.cos(theta), 0, np.sin(theta)],
                                        [0, 1, 0],
                                        [-np.sin(theta), 0, np.cos(theta)]])
        elif rotation_axis == 2:  # y
            rotation_matrix = np.array([[np.cos(theta), -np.sin(theta), 0],
                                        [np.sin(theta), np.cos(theta), 0],
                                        [0, 0, 1]])
    elif rank == 4:
        if rotation_axis == 0:  # z
            rotation_matrix = np.array([[1, 0, 0, 0],
                                        [0, np.cos(theta), -np.sin(theta), 0],
                                        [0, np.sin(theta), np.cos(theta), 0],
                                        [0, 0, 0, 1]])

        elif rotation_axis == 1:  # x
            rotation_matrix = np.array([[np.cos(theta), 0, np.sin(theta), 0],
                                        [0, 1, 0, 0],
                                        [-np.sin(theta), 0, np.cos(theta), 0],
                                        [0, 0, 0, 1]])
        elif rotation_axis == 2:  # y
            rotation_matrix = np.array([[np.cos(theta), -np.sin(theta), 0, 0],
                                        [np.sin(theta), np.cos(theta), 0, 0],
                                        [0, 0, 1, 0],
                                        [0, 0, 0, 1]])
    else:
        raise ValueError('Rotation matrix only support 3D and 4D tensor')

    return rotation_matrix


def get_shift_matrix(shift, rank=3):
    r"""
    Return the shift matrix in affine transform

    .. math::
        \begin{pmatrix}1&0&offset_x\\
                0&1&offset_y\\
                0&0&1
        \end{pmatrix}


    Parameters
    ----------
    shift : int,
        number of pixel to shift
    rank : int, optional
        the rank of the matrix, by default 3

    Returns
    -------
    np.array
        the shift matrix
    """
    shift_matrix = np.identity(rank)

    shift_matrix[:-1, -1] = shift

    return shift_matrix


def get_zoom_matrix(zoom_factor, rank=3):
    zoom_matrix = np.zeros((rank, rank))

    np.fill_diagonal(zoom_matrix, [zoom_factor for _ in range(rank-1)] + [1])

    return zoom_matrix


def transform_matrix_offset_center(matrix, width, height, depth=None):
    if matrix.shape[0] == 4 and depth is None:
        raise ValueError('4D tensor requires a depth value')

    z = width / 2 - 0.5
    x = height / 2 - 0.5

    if matrix.shape[0] == 3:
        offset_matrix = np.array([[1, 0, z],
                                  [0, 1, x],
                                  [0, 0, 1]])
        reset_matrix = np.array([[1, 0, -z],
                                 [0, 1, -x],
                                 [0, 0, 1]])

    elif matrix.shape[0] == 4:
        y = depth / 2 - 0.5
        offset_matrix = np.array([[1, 0, 0, z],
                                  [0, 1, 0, x],
                                  [0, 0, 1, y],
                                  [0, 0, 0, 1]])
        reset_matrix = np.array([[1, 0, 0, -z],
                                 [0, 1, 0, -x],
                                 [0, 0, 1, -y],
                                 [0, 0, 0, 1]])

    return np.dot(np.dot(offset_matrix, matrix), reset_matrix)


def affine_transform_matrix(rotation_axis=2, theta=0, rank=3,
                            zoom_factor=1, **kwargs):
    transform_matrix = None

    if theta:
        rotation_matrix = get_rotation_matrix(rotation_axis, theta, rank)
        transform_matrix = rotation_matrix

    # if shift:
    #     shift_matrix = get_shift_matrix(shift, rank)

    #     if transform_matrix is None:
    #         transform_matrix = shift_matrix
    #     else:
    #         transform_matrix = np.dot(transform_matrix, shift_matrix)

    if zoom_factor != 1:
        zoom_matrix = get_zoom_matrix(1/zoom_factor, rank)

        if transform_matrix is None:
            transform_matrix = zoom_matrix
        else:
            transform_matrix = np.dot(transform_matrix, zoom_matrix)

    return transform_matrix


def apply_affine_transform(image, mode='constant', cval=0, **kwargs):
    if not 3 <= image.ndim <= 4:
        raise ValueError(
            f'Not support affine transform for tensor of rank {image.ndim}')

    transform_matrix = affine_transform_matrix(rank=image.ndim, **kwargs)

    if transform_matrix is not None:
        offset_matrix = transform_matrix_offset_center(
            transform_matrix, *image.shape[:-1])

        offset = offset_matrix[..., -1]
        offset[-1] = 0
    else:
        transform_matrix = np.identity(image.ndim)
        offset = np.zeros(image.ndim)

    if kwargs.get('shift') is not None:
        offset[:-1] = offset[:-1] + np.array(kwargs['shift'])

    return ndimage.affine_transform(
        image, transform_matrix, offset, mode=mode, cval=cval)


def apply_flip(image, axis):
    image = np.array(image)
    if '__iter__' not in dir(axis):
        axis = [axis]
    if 0 in axis:
        image = image[::-1]

    if 1 in axis:
        image = image[:, ::-1]

    if 2 in axis:
        image = image[:, :, ::-1]

    return image



import numpy as np
from itertools import product


def get_patch_indice(shape, patch_size, overlap):
    if len(patch_size) != len(shape):
        raise ValueError

    res = []
    for length, size in zip(shape, patch_size):
        # the lower index will increase with step = size * (1-overlap)
        indice = np.arange(0, length - size, int(size*(1-overlap)))
        # add the last index, which may not follow the overlap rule
        if length - size != indice[-1]:
            indice = np.append(indice, length - size)

        res.append(indice)

    # return the combination of all dimensions
    return list(product(*res))


def get_patch_indice_all(images, patch_size, overlap):
    res = []
    for i, img in enumerate(images):
        indice = get_patch_indice(img.shape[:-1], patch_size, overlap)
        res.extend(product([i], indice))

    return res


def check_drop(images, patch_indice, patch_size,
               drop_fraction, check_drop_channel=None):
    if check_drop_channel is None:
        # all channel are checked
        check_images = images.copy()
        pixel_sum = np.prod(patch_size) * images[0].shape[-1]

        # all images are of same size, get the whole min
        if images.dtype is not object:
            imin = check_images.min()
    else:  # Only specific channels are checked
        pixel_sum = np.prod(patch_size)
        # all images are of same size, get the whole min, take the last axis
        if images.dtype is not object:
            # all images are
            check_images = images[..., check_drop_channel]
            imin = check_images.min()
        else:
            # create the new image one by one
            check_images = np.array([img[..., check_drop_channel]
                                     for img in images], dtype=object)

    res = []

    for i, indice in patch_indice:
        # imin has to be recalculated for each image with different sizes
        if images.dtype is object:
            imin = check_images[i].min()

        if len(indice) == 2:
            x, y = indice
            w, h = patch_size
            check_data = check_images[i][x:x+w, y:y+h].flatten()
            passed = len(check_data[check_data > imin]
                         ) / pixel_sum > drop_fraction
        elif len(indice) == 3:
            x, y, z = indice
            w, h, d = patch_size
            check_data = check_images[i][x:x+w, y:y+h, z:z+d].flatten()

            passed = len(check_data[check_data > imin]
                         ) / pixel_sum > drop_fraction

        res.append(passed)

    return res


def get_patches(images, target=None, patch_indice=None, patch_size=None,
                stratified=True, batch_size=None,
                drop_fraction=0.1, check_drop_channel=None):
    try:
        images = np.array(images)
    except Exception:  # compatity check  # pragma: no cover
        images = np.array(images, dtype=object)

    if images.dtype == object:  # pragma: no cover
        raise NotImplementedError(
            'Cannot handle images with different sizes.'
            'Consider using get_patches_different_images')

    if target is not None:
        try:
            target = np.array(target)
        except Exception:  # compatity check  # pragma: no cover
            target = np.array(target, dtype=object)
        if not images.shape[1:-1] == target.shape[1:-1]:
            raise ValueError(
                'Image and target shape are mismatched ({} != {})'.format(
                    images.shape[:-1], target.shape[:-1]))

    if patch_indice is None or patch_size is None:
        raise ValueError('patch_indice and patch_size are required.')

    pixel_sum = np.prod(patch_size)

    # create indice for each images
    # [(0, (x,y,z)), (0, (x',y',z')), (1, (x,y,z)), (1, (x',y',z'))]

    patch_indice = np.array(
        list((product(np.arange(len(images)), patch_indice))), dtype=object)

    if drop_fraction > 0:
        check_drop_list = check_drop(images, patch_indice, patch_size,
                                     drop_fraction, check_drop_channel)

        patch_indice = patch_indice[check_drop_list]

    patch_img = np.zeros((len(patch_indice), *patch_size, images.shape[-1]))

    if target is not None:
        patch_label = np.zeros((len(patch_indice), *patch_size, 1))
        if stratified:
            label_percentage = np.zeros(len(patch_indice))
    for i, (im_i, indice) in enumerate(patch_indice):
        if len(indice) == 2:
            x, y = indice
            w, h = patch_size
            patch_img[i] = images[im_i][x:x+w, y:y+h]

            if target is not None:
                patch_label[i] = target[im_i][x:x+w, y:y+h]
                if stratified:
                    label_percentage[i] = patch_label[i].sum() / pixel_sum
        elif len(indice) == 3:
            x, y, z = indice
            w, h, d = patch_size

            patch_img[i] = images[im_i][x:x+w, y:y+h, z:z+d]
            if target is not None:
                patch_label[i] = target[im_i][x:x+w, y:y+h, z:z+d]
                if stratified:
                    label_percentage[i] = patch_label[i].sum() / pixel_sum

    if target is None:
        return patch_img
    else:
        if stratified:
            new_index = np.arange(len(patch_indice))
            if batch_size is None or len(patch_indice) < batch_size:
                np.random.shuffle(new_index)
            else:
                new_index = get_stratified_index(label_percentage, batch_size)
            patch_img = patch_img[new_index]
            patch_label = patch_label[new_index]

        return patch_img, patch_label


def get_patches_different_images(images, target=None, patch_indice=None,
                                 patch_size=None, overlap=0,
                                 stratified=True,
                                 drop_fraction=0.1,
                                 check_drop_channel=None):  # pragma: no cover
    images = np.array(images, dtype=object)
    target = np.array(target, dtype=object)

    indice = get_patch_indice_all(images, patch_size, overlap)

    raise NotImplementedError


def get_stratified_index(values, sample_size):
    median = np.median(values)

    index = np.arange(len(values))
    positive_index = index[values > median]
    negative_index = index[values <= median]

    np.random.shuffle(positive_index)
    np.random.shuffle(negative_index)

    split_num = int(np.ceil(len(values) / sample_size))
    pos_num = len(positive_index) // split_num
    step = 1

    inners = [[] for _ in range(split_num)]
    # cannot evenly distribute in each batch
    if pos_num == 0:
        step = int(np.ceil(split_num / len(positive_index)))

    while len(positive_index) > 0:
        for i in range(0, split_num, step):
            if len(positive_index) == 0:
                break
            inners[i].append(positive_index[-1])
            positive_index = positive_index[:-1]

    for inner in inners:
        while len(inner) < sample_size and len(negative_index) > 0:
            inner.append(negative_index[-1])
            negative_index = negative_index[:-1]
        np.random.shuffle(inner)

    # if len(np.concatenate(inners)) != len(values):
    #     raise RuntimeError(
    #         'Something wrong happened. Please contact the developer')

    # if len(negative_index) > 0 or len(positive_index) > 0:
    #     raise RuntimeError(
    #         'Something wrong happened. Please contact the developer')

    return np.concatenate(inners)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.2"


from .affine_transform import *
from .point_operation import *
from .filters import *
from .patch_sliding import *
from .augmentation import *



import pytest
import numpy as np
from deoxys_image import apply_random_gaussian_blur


def test_gaussian_blue():
    images = np.random.normal(0, 1, size=(30, 5, 5, 3))

    res = apply_random_gaussian_blur(images, low=0.6, high=1.3)
    assert not np.allclose(res, images)

    res = apply_random_gaussian_blur(images, low=0.6, high=1.3, channel=0)
    assert not np.allclose(res[..., 0], images[..., 0])
    assert np.allclose(res[..., 1:], images[..., 1:])

    res = apply_random_gaussian_blur(images, low=0.6, high=1.3, channel=[0, 1])
    assert not np.allclose(res[..., 0], images[..., 0])
    assert not np.allclose(res[..., 1], images[..., 1])
    assert np.allclose(res[..., 2], images[..., 2])



import numpy as np


def change_brightness(image, factor, channel=None):
    """
    Change brightness of an images by adding a constant to all pixels
    in the image so that `max(new_image) = factor * max(image)`

    Parameters
    ----------
    image : np.array
        the image
    factor : float
        change factor for brightness. Factor > 1 means the image will
        get brighter
    channel : int or list, optional
        channel or channels to be adjusted, by default None,
        which means all channels will be adjusted

    Returns
    -------
    np.array
        image with changes in brightness
    """
    image = np.array(image)

    if channel is None:
        imin = image.min()
        imax = image.max()

        change_const = imax * (factor - 1)
        return (image + change_const).clip(imin, imax)
    else:
        if '__iter__' not in dir(channel):
            channel = [channel]

        for c in channel:
            imin = image[..., c].min()
            imax = image[..., c].max()

            change_const = imax * (factor - 1)
            image[..., c] = (image[..., c] + change_const).clip(imin, imax)

        return image


def change_contrast(image, factor, channel=None):
    """
    Change contrast of an images by a factor

    Parameters
    ----------
    image : np.array
        the image
    factor : float
        change factor for contrast. Factor > 1 increases the contrast
    channel : int or list, optional
        channel or channels to be adjusted, by default None,
        which means all channels will be adjusted

    Returns
    -------
    np.array
        image with changes in contrast
    """
    image = np.array(image)

    if channel is None:
        imin = image.min()
        imax = image.max()
        imean = image.mean()

        return ((image - imean) * factor + imean).clip(imin, imax)
    else:
        if '__iter__' not in dir(channel):
            channel = [channel]

        for c in channel:
            imin = image[..., c].min()
            imax = image[..., c].max()
            imean = image[..., c].mean()

            image[..., c] = ((image[..., c] - imean) *
                             factor + imean).clip(imin, imax)

        return image


def apply_random_contrast(images, low=0.6, high=1.4, channel=None):
    images = np.array(images)

    change_factor = np.random.sample(images.shape[0])*(high - low) + low

    for i in range(images.shape[0]):
        images[i] = change_contrast(
            images[i], factor=change_factor[i], channel=channel)

    return images


def apply_random_brightness(images, low=0.7, high=1.3, channel=None):
    images = np.array(images)
    change_factor = np.random.sample(images.shape[0])*(high - low) + low

    for i in range(images.shape[0]):
        images[i] = change_brightness(
            images[i], factor=change_factor[i], channel=channel)

    return images


def gaussian_noise(image, noise_var=0.1, channel=None):
    """
    Apply gaussian noise of variance `noise_var` to a single image

    Parameters
    ----------
    image : np.array
        the image
    noise_var : float, optional
        the variance of the noise value, by default 0.1
    channel : int or list, optional
        channel or channels to be added with random noise, by default None,
        which means all channels are added with random noise
    Returns
    -------
    np.array
        images with noise
    """
    image = np.array(image)

    if channel is None:
        return image + np.random.normal(0, noise_var, size=image.shape)
    else:
        if '__iter__' in dir(channel):
            shape = list(image.shape[:-1]) + [len(channel)]
        else:
            shape = list(image.shape[:-1])

        image[..., channel] = image[..., channel] + \
            np.random.normal(0, noise_var, size=shape)

        return image


def apply_gaussian_noise(images, noise_var=0.1,
                         channel=None):  # pragma: no cover
    """
    Apply gaussian noise of variance `noise_var`

    Parameters
    ----------
    images : np.array
        images
    noise_var : float, optional
        the variance of the noise value
    channel: int or list, optional
        channel or channels to be added with random noise, by default None,
        which means all channels are added with random noise

    Returns
    -------
    np.array
        images with noise
    """
    images = np.array(images)

    if channel is None:
        return images + np.random.normal(0, noise_var, size=images.shape)
    else:
        if '__iter__' not in dir(channel):
            shape = list(images.shape[:-1]) + [len(channel)]
        else:
            shape = list(images.shape[:-1])

        images[..., channel] = images[..., channel] + \
            np.random.normal(0, noise_var, size=shape)

        return images


def apply_random_gaussian_noise(images, variance=None, channel=None):
    """
    Apply gaussian noise with random noise variance in the range of `variance`

    Parameters
    ----------
    images : np.array
        images
    variance : tuple or float, optional
        range of the noise variance, by default (0, 0.1)
    channel: int or list, optional
        channel or channels to be added with random noise, by default None,
        which means all channels are added with random noise
    """
    images = np.array(images)

    if variance is None:
        variance = (0, 0.1)

    if '__iter__' not in dir(variance):
        variance = (0, variance)

    variances = np.random.sample(
        images.shape[0])*(variance[1] - variance[0]) + variance[0]
    # if channel is None:
    #     for i in range(images.shape[0]):
    #         images[i] = images[i] + \
    #             np.random.normal(0.0, variances[i], size=images.shape[1:])
    # else:
    #     if '__iter__' not in dir(channel):
    #         shape = list(images.shape[1:-1]) + [len(channel)]
    #     else:
    #         shape = list(images.shape[1:-1])

    #     for i in range(images.shape[0]):
    #         images[i][..., channel] = images[i][..., channel] + \
    #             np.random.normal(0.0, variances[i], size=shape)

    # return images

    for i in range(images.shape[0]):
        images[i] = gaussian_noise(images[i], variances[i], channel=channel)

    return images


def normalize(images, vmin=None, vmax=None):
    images = np.array(images)
    n_channel = images.shape[-1]

    if vmin is None:
        vmin = np.array([images[..., i].min() for i in range(n_channel)])
    if vmax is None:
        vmax = np.array([images[..., i].max() for i in range(n_channel)])

    if '__iter__' in dir(vmin):
        vmin = np.array(vmin)
        if len(vmin) < n_channel:
            template = np.array([images[..., i].min()
                                 for i in range(n_channel)])
            template[:len(vmin)] = vmin
            vmin = template

    if '__iter__' in dir(vmax):
        vmax = np.array(vmax)

        if len(vmax) < n_channel:
            template = np.array([images[..., i].max()
                                 for i in range(n_channel)])
            template[:len(vmax)] = vmax
            vmax = template

    transformed_images = (images - vmin) / (vmax - vmin)
    transformed_images = transformed_images.clip(0, 1)

    return transformed_images



import pytest
import numpy as np

from deoxys_image import ImageAugmentation, get_random_affine_params


def test_get_random_affine_params():
    for _ in range(100):
        theta, zoom, shift = get_random_affine_params(
            (-4, 4), 0.2, (0.8, 1.2), 0.2, (np.zeros(2), np.zeros(2)), 0.1)

        assert -4 <= theta < 4
        assert 0.8 <= zoom < 1.2
        assert np.all(shift == 0)
        assert len(shift) == 2

    no_changes = 0
    for _ in range(100):
        theta, zoom, shift = get_random_affine_params(
            (-4, 4), 0.2,
            (0.8, 1.2), 0.2,
            (np.array([0, -4]), np.array([0, 4])), 0.1)

        assert -4 <= theta < 4
        assert 0.8 <= zoom < 1.2
        assert shift[0] == 0
        assert -4 <= shift[1] < 4
        assert len(shift) == 2

        if theta == 0 and zoom == 1 and np.all(shift == 0):
            no_changes += 1

    assert no_changes > 50


def test_augmentation_single():
    images = np.random.random((10, 3, 3, 2))
    targets = np.random.random((10, 3, 3, 1))

    aug = ImageAugmentation(rank=3, rotation_range=20, rotation_axis=2)
    aug.transform(images)
    aug.transform(images, targets)

    aug = ImageAugmentation(rank=3, zoom_range=1.5)
    aug.transform(images)

    aug = ImageAugmentation(rank=3, zoom_range=0.5)
    aug.transform(images, targets)

    aug = ImageAugmentation(rank=3, zoom_range=(0.8, 1.2))
    aug.transform(images)

    aug = ImageAugmentation(rank=3, shift_range=(0, 20))
    aug.transform(images)

    aug = ImageAugmentation(rank=3, shift_range=30)
    aug.transform(images, targets)

    aug = ImageAugmentation(rank=3, flip_axis=0)
    aug.transform(images, targets)

    aug = ImageAugmentation(rank=3, flip_axis=[0, 1])
    aug.transform(images)

    aug = ImageAugmentation(rank=3, brightness_range=0.8)
    aug.transform(images)

    aug = ImageAugmentation(rank=3, brightness_range=1.2)
    aug.transform(images)

    aug = ImageAugmentation(rank=3, brightness_range=(0.8, 1.2))
    aug.transform(images)

    aug = ImageAugmentation(rank=4, contrast_range=0.8)
    aug.transform(images)

    aug = ImageAugmentation(rank=4, contrast_range=1.2)
    aug.transform(images)

    aug = ImageAugmentation(rank=4, contrast_range=(0.8, 1.2))
    aug.transform(images)

    aug = ImageAugmentation(rank=4, noise_variance=0.8)
    aug.transform(images)

    aug = ImageAugmentation(rank=4, noise_variance=1.2)
    aug.transform(images)

    aug = ImageAugmentation(rank=4, noise_variance=(0.8, 1.2))
    aug.transform(images)

    aug = ImageAugmentation(rank=4, blur_range=0.8)
    aug.transform(images)

    aug = ImageAugmentation(rank=4, blur_range=1.2)
    aug.transform(images)

    aug = ImageAugmentation(rank=4, blur_range=(0.8, 1.2))
    aug.transform(images)

    with pytest.raises(ValueError):
        aug = ImageAugmentation(rank=3, shift_range=(0, -20))



import numpy as np

from .point_operation import gaussian_noise, change_brightness, change_contrast
from .filters import gaussian_blur
from .affine_transform import apply_affine_transform, apply_flip


class ImageAugmentation():
    r"""
        Apply transformation in 2d and 3d image (and mask label)
        for augmentation

        Parameters
        ----------
        rank : int
            rank of a single image (2d: 3, 3d: 4)
        rotation_range : int, optional
            range of the angle rotation, in degree, by default 0 (no rotation)
        rotation_axis : int, optional
            the axis of one image to apply rotation,
            by default 0
        rotation_chance : float, optional
            probability to apply rotation transformation to an image,
            by default 0.2
        zoom_range : float, list, tuple optional
            the range of zooming, zooming in when the number is less than 1,
            and zoom out when the number if larger than 1.
            If a `float`, then it is the range between that number and 1,
            by default 1 (no zooming)
        zoom_chance : float, optional
            probability to apply zoom transformation to an image,
            by default 0.2
        shift_range : tuple or list, optional
            the range of translation in each axis, by default None (no shifts)
        shift_chance : float, optional
            probability to apply translation transformation to an image,
            by default 0.1
        flip_axis : int, tuple, list, optional
            flip by one or more axis (in the single image),
            by default None (no flipping)
        brightness_range : int, tuple, list, optional
            range of the brightness portion,
            based on the max intensity value of each channel.
            For example, when the max intensity value of one channel is 1.0,
            and the brightness is chaned by 1.2, then every pixel in that
            channel will increase the intensity value by 0.2.

            .. math:: 0.2 = 1.0 \cdot (1.2 - 1)

            By default 1 (no changes in brightness)
        brightness_channel : int, tuple, list, optional
            the channel(s) to apply changes in brightness,
            by default None (apply to all channels)
        brightness_chance : float, optional
            probability to apply brightness change transform to an image,
            by default 0.1
        contrast_range : int, tuple, list, optional
            range of the contrast portion,
            (the history range is scaled up or down).
            By default 1 (no changes in contrast)
        contrast_channel : int, tuple, list, optional
            the channel(s) to apply changes in contrast,
            by default None (apply to all channels)
        contrast_chance : float, optional
            probability to apply contrast change transform to an image,
            by default 0.1
        noise_variance : int, tuple, list, optional
            range of the noise variance
            when adding Gaussian noise to the image,
            by default 0 (no adding noise)
        noise_channel : int, tuple, list, optional
            the channel(s) to apply Gaussian noise,
            by default None (apply to all channels)
        noise_chance : float, optional
            probability to apply gaussian noise to an image,
            by default 0.1
        blur_range : int, tuple, list, optional
            range of the blur sigma
            when applying the Gaussian filter to the image,
            by default 0 (no blur)
        blur_channel :int, tuple, list, optional
            the channel(s) to apply Gaussian blur,
            by default None (apply to all channels)
        blur_chance : float, optional
            probability to apply gaussian blur to an image,
            by default 0.1
        fill_mode : str, optional
            the fill mode in affine transformation
            (rotation, zooming, shifting / translation),
            one of {'reflect', 'constant', 'nearest', 'mirror', 'wrap'},
            by default 'constant'
        cval : int, optional
            When rotation, or zooming, or shifting is applied to the image,
            `cval` is the value to fill past edges of input
            if `fill_mode` is 'constant'.
            By default 0
        """

    def __init__(self, rank,
                 rotation_range=0, rotation_axis=0, rotation_chance=0.2,
                 zoom_range=1, zoom_chance=0.2,
                 shift_range=None, shift_chance=0.1,
                 flip_axis=None,
                 brightness_range=1, brightness_channel=None,
                 brightness_chance=0.1,
                 contrast_range=1, contrast_channel=None,
                 contrast_chance=0.1,
                 noise_variance=0, noise_channel=None,
                 noise_chance=0.1,
                 blur_range=0, blur_channel=None, blur_chance=0.1,
                 fill_mode='constant', cval=0):
        """
        Apply transformation in 2d and 3d image for augmentation
        """
        # check if perform affine transform
        self.affine_transform = rotation_range > 0 or \
            zoom_range != 1 or shift_range is not None

        # get the low high limit of the range
        if self.affine_transform:
            # 2d: rank=3; 3d: rank=4
            self.rotation_range, self.zoom_range, \
                self.shift_range = get_range_affine_transform(
                    rank, rotation_range, zoom_range, shift_range)

            self.rotation_axis = rotation_axis
            self.rotation_chance = rotation_chance

            self.zoom_chance = zoom_chance
            self.shift_chance = shift_chance

            self.fill_mode = fill_mode
            self.cval = cval

        # flip channel in the form of iterator
        if flip_axis is not None and '__iter__' not in dir(flip_axis):
            self.flip_axis = [flip_axis]
        else:
            self.flip_axis = flip_axis

        self.brightness_range = brightness_range

        # low, high of brightness
        if brightness_range != 1:
            self.brightness_range = get_range_value(
                brightness_range,
                default_val=1)
            self.brightness_chance = brightness_chance
            self.brightness_channel = brightness_channel
        else:
            self.brightness_range = brightness_range

        # low, high of contrast
        if contrast_range != 1:
            self.contrast_range = get_range_value(
                contrast_range,
                default_val=1)
            self.contrast_chance = contrast_chance
            self.contrast_channel = contrast_channel
        else:
            self.contrast_range = contrast_range

        # low, high of noise
        if noise_variance != 0:
            self.noise_variance = get_range_value(
                noise_variance,
                default_val=0)
            self.noise_chance = noise_chance
            self.noise_channel = noise_channel
        else:
            self.noise_variance = noise_variance

        # low, high of blur sigma
        if blur_range != 0:
            self.blur_range = get_range_value(
                blur_range,
                default_val=0)
            self.blur_chance = blur_chance
            self.blur_channel = blur_channel
        else:
            self.blur_range = blur_range

    def transform(self, images, targets=None):
        """
        Apply augmentation to a batch of images

        Parameters
        ----------
        images : np.array
            the image batch
        targets : np.array, optional
            the target batch, by default None

        Returns
        -------
        np.array
            the transformed images batch (and target)
        """
        # copy to another version
        transformed_images = images.copy()
        if targets is not None:
            transformed_targets = targets.copy()

        # loop through
        for i in range(len(images)):
            # apply affine transform if possible
            if self.affine_transform:
                theta, zoom_factor, shift = get_random_affine_params(
                    rotation_range=self.rotation_range,
                    rotation_chance=self.rotation_chance,
                    zoom_range=self.zoom_range,
                    zoom_chance=self.zoom_chance,
                    shift_range=self.shift_range,
                    shift_chance=self.shift_chance)
                # Only apply affine transform when needed
                if theta != 0 or zoom_factor != 1 or not np.all(shift == 0):
                    # After affine transform, the pixel intensity may change
                    # the image should clip back to original range
                    reduced_ax = tuple(
                        range(len(transformed_images[i].shape) - 1))
                    vmin = transformed_images[i].min(axis=reduced_ax)
                    vmax = transformed_images[i].max(axis=reduced_ax)

                    transformed_images[i] = apply_affine_transform(
                        transformed_images[i],
                        mode=self.fill_mode, cval=self.cval,
                        theta=theta, rotation_axis=self.rotation_axis,
                        zoom_factor=zoom_factor,
                        shift=shift).clip(vmin, vmax)

                    if targets is not None:
                        transformed_targets[i] = apply_affine_transform(
                            transformed_targets[i],
                            mode=self.fill_mode, cval=self.cval,
                            theta=theta, rotation_axis=self.rotation_axis,
                            zoom_factor=zoom_factor,
                            shift=shift)
                        # round the target label back to integer
                        transformed_targets[i] = np.rint(
                            transformed_targets[i])

            # flip image
            if self.flip_axis is not None:
                actual_flip_axis = []
                for channel in self.flip_axis:
                    if np.random.uniform() < 0.5:
                        actual_flip_axis.append(channel)

                if len(actual_flip_axis) > 0:
                    transformed_images[i] = apply_flip(
                        transformed_images[i], actual_flip_axis)

                    if targets is not None:
                        transformed_targets[i] = apply_flip(
                            transformed_targets[i], actual_flip_axis)

            # brightness
            if self.brightness_range != 1 and \
                    np.random.uniform() < self.brightness_chance:
                transformed_images[i] = change_brightness(
                    transformed_images[i],
                    np.random.uniform(*self.brightness_range),
                    channel=self.brightness_channel)

            # contrast
            if self.contrast_range != 1 and \
                    np.random.uniform() < self.contrast_chance:
                transformed_images[i] = change_contrast(
                    transformed_images[i],
                    np.random.uniform(*self.contrast_range),
                    channel=self.contrast_channel)

            # gaussian noise
            if self.noise_variance != 0 and \
                    np.random.uniform() < self.noise_chance:
                transformed_images[i] = gaussian_noise(
                    transformed_images[i],
                    np.random.uniform(*self.noise_variance),
                    channel=self.noise_channel)

            # gaussian blur
            if self.blur_range != 0 and np.random.uniform() < self.blur_chance:
                transformed_images[i] = gaussian_blur(
                    transformed_images[i],
                    np.random.uniform(*self.blur_range),
                    channel=self.blur_channel)

        if targets is None:
            return transformed_images
        else:
            return transformed_images, transformed_targets


def get_range_value(value, default_val=1):
    if '__iter__' in dir(value):
        low, high = value
    elif value < default_val:
        low, high = value, default_val
    else:
        low, high = default_val, value

    return low, high


def get_range_affine_transform(rank, rotation_range,
                               zoom_range,
                               shift_range):

    if shift_range is None:
        shift_low = np.zeros(rank-1)
        shift_high = np.zeros(rank-1)
    else:
        if '__iter__' not in dir(shift_range):
            shift_range = np.array([shift_range] * (rank-1))
        else:
            shift_range = np.array(shift_range)

        if not np.all(shift_range >= 0):
            raise ValueError('All elements in shift range should >= 0')

        shift_low = -shift_range
        shift_high = shift_range

    rotation_low = -rotation_range
    rotation_high = rotation_range

    if '__iter__' in dir(zoom_range):
        zoom_low, zoom_high = zoom_range
    elif zoom_range < 1:
        zoom_low, zoom_high = zoom_range, 1
    else:
        zoom_low, zoom_high = 1, zoom_range

    return (rotation_low, rotation_high), (zoom_low, zoom_high), \
        (shift_low, shift_high)


def get_random_affine_params(rotation_range,
                             rotation_chance,
                             zoom_range, zoom_chance,
                             shift_range, shift_chance):
    if np.random.uniform() < rotation_chance:
        theta = np.random.uniform(*rotation_range)
    else:
        theta = 0

    if np.random.uniform() < zoom_chance:
        zoom_factor = np.random.uniform(*zoom_range)
    else:
        zoom_factor = 1

    if np.random.uniform() < shift_chance:
        shift_factor = np.random.uniform(*shift_range)
    else:
        shift_factor = np.zeros(len(shift_range[0]))

    return theta, zoom_factor, shift_factor


def apply_augmentation(images, targets=None,
                       rotation_range=0, rotation_axis=0, rotation_chance=0.2,
                       zoom_range=1, zoom_chance=0.2,
                       shift_range=None, shift_chance=0.1,
                       flip_axis=None,
                       brightness_range=1, brightness_channel=None,
                       brightness_chance=0.1,
                       contrast_range=1, contrast_channel=None,
                       contrast_chance=0.1,
                       noise_variance=0, noise_channel=None,
                       noise_chance=0.1,
                       blur_range=0, blur_channel=None, blur_chance=0.1,
                       fill_mode='constant', cval=0):  # pragma: no cover
    # copy to another version
    transformed_images = images.copy()
    if targets is not None:
        transformed_targets = targets.copy()

    # check if perform affine transform
    affine_transform = rotation_range > 0 or \
        zoom_range != 1 or shift_range is not None

    # get the low high limit of the range
    if affine_transform:
        rotation, zoom, shift_var = get_range_affine_transform(
            images.ndim - 1, rotation_range, zoom_range, shift_range)

    # flip channel in the form of iterator
    if flip_axis is not None and '__iter__' not in dir(flip_axis):
        flip_axis = [flip_axis]

    for i, image in enumerate(images):
        if affine_transform:
            theta, zoom_factor, shift = get_random_affine_params(
                rotation_range=rotation, rotation_chance=rotation_chance,
                zoom_range=zoom, zoom_chance=zoom_chance,
                shift_range=shift_var, shift_chance=shift_chance)
            # Only apply affine transform when needed
            if theta != 0 or zoom_factor != 1 or not np.all(shift == 0):
                transformed_images[i] = apply_affine_transform(
                    transformed_images[i],
                    mode=fill_mode, cval=cval,
                    theta=theta, rotation_axis=rotation_axis,
                    zoom_factor=zoom_factor,
                    shift=shift)

                if targets is not None:
                    transformed_targets[i] = apply_affine_transform(
                        transformed_targets[i],
                        mode=fill_mode, cval=cval,
                        theta=theta, rotation_axis=rotation_axis,
                        zoom_factor=zoom_factor,
                        shift=shift)

        if flip_axis is not None:
            actual_flip_axis = []
            for channel in flip_axis:
                if np.random.uniform() < 0.5:
                    actual_flip_axis.append(channel)

            if len(actual_flip_axis) > 0:
                transformed_images[i] = apply_flip(
                    transformed_images[i], actual_flip_axis)

                if targets is not None:
                    transformed_targets[i] = apply_flip(
                        transformed_targets[i], actual_flip_axis)

        if brightness_range != 1 and np.random.uniform() < brightness_chance:
            bright_low, bright_high = get_range_value(brightness_range,
                                                      default_val=1)

            transformed_images[i] = change_brightness(
                transformed_images[i],
                np.random.uniform(bright_low, bright_high),
                channel=brightness_channel)

        if contrast_range != 1 and np.random.uniform() < contrast_chance:
            contrast_low, contrast_high = get_range_value(contrast_range,
                                                          default_val=1)

            transformed_images[i] = change_contrast(
                transformed_images[i],
                np.random.uniform(contrast_low, contrast_high),
                channel=contrast_channel)

        if noise_variance > 0 and np.random.uniform() < noise_chance:
            noise_low, noise_high = get_range_value(noise_variance,
                                                    default_val=0)

            transformed_images[i] = gaussian_noise(
                transformed_images[i],
                np.random.uniform(noise_low, noise_high),
                channel=noise_channel)

        if blur_range != 0 and np.random.uniform() < blur_chance:
            blur_low, blur_high = get_range_value(blur_range,
                                                  default_val=0)

            transformed_images[i] = gaussian_blur(
                transformed_images[i],
                np.random.uniform(blur_low, blur_high),
                channel=blur_channel)

    if targets is None:
        return transformed_images
    else:
        return transformed_images, transformed_targets



from scipy.ndimage import gaussian_filter
import numpy as np


def gaussian_blur(image, sigma=0.4, channel=None):
    image = np.array(image)

    if channel is None:
        channel = [i for i in range(image.shape[-1])]

    if '__iter__' not in dir(channel):
        channel = [channel]
    for c in channel:
        image[..., c] = gaussian_filter(
            image[..., c], sigma=sigma, order=0)
    return image


def apply_random_gaussian_blur(images, low=0.5, high=1.5, channel=None):
    images = np.array(images)
    sigma = np.random.sample(images.shape[0])*(high - low) + low

    for i in range(images.shape[0]):
        images[i] = gaussian_blur(
            images[i], sigma=sigma[i], channel=channel)

    return images



import pytest
import numpy as np
from deoxys_image import normalize
from deoxys_image import apply_random_brightness, apply_random_contrast
from deoxys_image import apply_random_gaussian_noise


def test_normalize_all_channel():
    base_data = np.array([np.arange(30) for _ in range(5)])

    normalized_image = np.zeros(30)
    normalized_image[20:] = 1
    normalized_image[10:20] = np.arange(10) / 10

    normalize_data = np.array([normalized_image for _ in range(5)])

    images = np.zeros((5, 5, 6, 2))
    images[..., 0] = base_data.reshape(5, 5, 6)
    images[..., 1] = base_data.reshape(5, 5, 6)

    expected = np.zeros((5, 5, 6, 2))
    expected[..., 0] = normalize_data.reshape(5, 5, 6)
    expected[..., 1] = normalize_data.reshape(5, 5, 6)

    res = normalize(images, 10, 20)

    assert np.allclose(res, expected)

    res = normalize(images)
    assert np.allclose(res, images/29)


def test_normalize_per_channel():
    base_data = np.array([np.arange(30) for _ in range(5)])

    normalized_image = np.zeros(30)
    normalized_image[20:] = 1
    normalized_image[10:20] = np.arange(10) / 10

    normalize_data = np.array([normalized_image for _ in range(5)])

    images = np.zeros((5, 5, 6, 3))
    images[..., 0] = base_data.reshape(5, 5, 6)
    images[..., 1] = base_data.reshape(5, 5, 6)
    images[..., 2] = base_data.reshape(5, 5, 6)

    expected = np.zeros((5, 5, 6, 3))
    expected[..., 0] = normalize_data.reshape(5, 5, 6)
    expected[..., 1] = normalize_data.reshape(5, 5, 6)
    expected[..., 2] = base_data.reshape(5, 5, 6) / 29

    res = normalize(images, [10, 10], [20, 20])

    assert np.allclose(res, expected)

    res = normalize(images, [10, 20], [20, 30])

    normalized_image = np.zeros(30)
    normalized_image[20:] = np.arange(10) / 10
    normalize_data = np.array([normalized_image for _ in range(5)])
    expected[..., 1] = normalize_data.reshape(5, 5, 6)

    assert np.allclose(res, expected)


def test_brightness_constant_all_channel():
    base_data = np.array([np.arange(1, 31) / 30 for _ in range(5)])

    images = np.zeros((5, 5, 6, 2))
    images[..., 0] = base_data.reshape(5, 5, 6)
    images[..., 1] = base_data.reshape(5, 5, 6)

    res = apply_random_brightness(images, low=1.5, high=1.5)

    assert np.allclose(res, (images + 0.5).clip(1/30, 1))

    res = apply_random_brightness(images, low=0.7, high=0.7)

    assert np.allclose(res, (images - 0.3).clip(1/30, 1))


def test_brightness_random_all_channel():
    base_data = np.array([np.arange(1, 31) / 30 for _ in range(30)])

    images = np.zeros((30, 5, 6, 2))
    images[..., 0] = base_data.reshape(30, 5, 6)
    images[..., 1] = base_data.reshape(30, 5, 6)

    res = apply_random_brightness(images, low=0.8, high=1.2)

    for i, img in enumerate(res):
        diff = img[2, 2, 0] - images[i, 2, 2, 0]
        assert -0.2 <= diff <= 0.2
        assert np.allclose(img, (images[i] + diff).clip(1/30, 1))


def test_brightness_random_per_channel():
    base_data = np.array([np.arange(1, 31) / 30 for _ in range(30)])

    images = np.zeros((30, 5, 6, 2))
    images[..., 0] = base_data.reshape(30, 5, 6)
    images[..., 1] = base_data.reshape(30, 5, 6)

    res = apply_random_brightness(images, low=0.8, high=1.2, channel=0)

    for i, img in enumerate(res):
        diff = img[2, 2] - images[i, 2, 2]
        assert -0.2 <= diff[0] <= 0.2
        assert np.allclose(img, (images[i] + diff).clip(1/30, 1))

    images = np.zeros((30, 5, 6, 3))
    images[..., 0] = base_data.reshape(30, 5, 6)
    images[..., 1] = base_data.reshape(30, 5, 6)
    images[..., 2] = base_data.reshape(30, 5, 6)

    res = apply_random_brightness(images, low=0.8, high=1.2, channel=[0, 1])

    for i, img in enumerate(res):
        diff = img[2, 2] - images[i, 2, 2]
        assert -0.2 <= diff[0] <= 0.2
        assert -0.2 <= diff[1] <= 0.2
        assert np.allclose(img, (images[i] + diff).clip(1/30, 1))


def test_contrast_constant_all_channel():
    base_data = np.array([np.arange(1, 31) / 30 for _ in range(5)])

    images = np.zeros((5, 5, 6, 2))
    images[..., 0] = base_data.reshape(5, 5, 6)
    images[..., 1] = base_data.reshape(5, 5, 6)

    res = apply_random_contrast(images, low=1.5, high=1.5)

    mn = images.mean()

    assert np.allclose(res, ((images - mn) * 1.5 + mn).clip(1/30, 1))

    res = apply_random_contrast(images, low=0.7, high=0.7)

    assert np.allclose(res, ((images - mn) * 0.7 + mn).clip(1/30, 1))


def test_contrast_random_all_channel():
    base_data = np.array([np.arange(1, 31) / 30 for _ in range(30)])

    images = np.zeros((30, 5, 6, 2))
    images[..., 0] = base_data.reshape(30, 5, 6)
    images[..., 1] = base_data.reshape(30, 5, 6)

    res = apply_random_contrast(images, low=0.8, high=1.2)
    mn = images.mean()

    for i, img in enumerate(res):
        diff = (img[2, 2, 0] - mn) / (images[i, 2, 2, 0] - mn)
        assert 0.8 <= diff <= 1.2
        assert np.allclose(img, ((images[i] - mn) * diff + mn).clip(1/30, 1))


def test_contrast_random_per_channel():
    base_data = np.array([np.arange(1, 31) / 30 for _ in range(30)])

    images = np.zeros((30, 5, 6, 2))
    images[..., 0] = base_data.reshape(30, 5, 6)
    images[..., 1] = base_data.reshape(30, 5, 6)

    res = apply_random_contrast(images, low=0.8, high=1.2, channel=0)
    mn = images.mean()

    for i, img in enumerate(res):
        diff = (img[2, 2] - mn) / (images[i, 2, 2] - mn)

        assert 0.8 <= diff[0] <= 1.2

        assert np.allclose(img, ((images[i] - mn) * diff + mn).clip(1/30, 1))

    images = np.zeros((30, 5, 6, 3))
    images[..., 0] = base_data.reshape(30, 5, 6)
    images[..., 1] = base_data.reshape(30, 5, 6)
    images[..., 2] = base_data.reshape(30, 5, 6)

    res = apply_random_contrast(
        images, low=0.8, high=1.2, channel=[0, 1])

    for i, img in enumerate(res):
        diff = (img[2, 2] - mn) / (images[i, 2, 2] - mn)

        assert 0.8 <= diff[0] <= 1.2
        assert 0.8 <= diff[1] <= 1.2
        assert np.allclose(img, ((images[i] - mn) * diff + mn).clip(1/30, 1))


def test_gaussian_noise():
    images = np.random.normal(0, 1, size=(30, 5, 5, 3))

    res = apply_random_gaussian_noise(images, (0.1, 0.3))
    assert not np.allclose(res, images)

    res = apply_random_gaussian_noise(images, 0.5, channel=0)
    assert not np.allclose(res[..., 0], images[..., 0])
    assert np.allclose(res[..., 1:], images[..., 1:])

    res = apply_random_gaussian_noise(images, 0.5, channel=[0, 1])
    assert not np.allclose(res[..., 0], images[..., 0])
    assert not np.allclose(res[..., 1], images[..., 1])
    assert np.allclose(res[..., 2], images[..., 2])



import pytest
import numpy as np
from itertools import product
from deoxys_image import get_patch_indice, get_patch_indice_all, check_drop
from deoxys_image import get_stratified_index, get_patches


def test_get_patch_indice_error():
    with pytest.raises(ValueError):
        get_patch_indice((2, 3, 4), (13, 13), 0)


def test_get_patch_indice():
    indice = get_patch_indice((20, 20, 20), (10, 10, 10), 0)

    expected = [(0, 0, 0), (0, 0, 10),
                (0, 10, 0), (0, 10, 10),
                (10, 0, 0), (10, 0, 10),
                (10, 10, 0), (10, 10, 10)]

    assert np.all(indice == expected)

    indice = get_patch_indice((20, 20, 20), (10, 10, 10), 0.5)

    expected = [(0, 0, 0), (0, 0, 5), (0, 0, 10),
                (0, 5, 0), (0, 5, 5), (0, 5, 10),
                (0, 10, 0), (0, 10, 5), (0, 10, 10),
                (5, 0, 0), (5, 0, 5), (5, 0, 10),
                (5, 5, 0), (5, 5, 5), (5, 5, 10),
                (5, 10, 0), (5, 10, 5), (5, 10, 10),
                (10, 0, 0), (10, 0, 5), (10, 0, 10),
                (10, 5, 0), (10, 5, 5), (10, 5, 10),
                (10, 10, 0), (10, 10, 5), (10, 10, 10)]

    assert np.all(indice == expected)

    indice = get_patch_indice((20, 20, 19), (10, 10, 10), 0.5)

    expected = [(0, 0, 0), (0, 0, 5), (0, 0, 9),
                (0, 5, 0), (0, 5, 5), (0, 5, 9),
                (0, 10, 0), (0, 10, 5), (0, 10, 9),
                (5, 0, 0), (5, 0, 5), (5, 0, 9),
                (5, 5, 0), (5, 5, 5), (5, 5, 9),
                (5, 10, 0), (5, 10, 5), (5, 10, 9),
                (10, 0, 0), (10, 0, 5), (10, 0, 9),
                (10, 5, 0), (10, 5, 5), (10, 5, 9),
                (10, 10, 0), (10, 10, 5), (10, 10, 9)]

    assert np.all(indice == expected)


def test_get_patch_indice_all():
    images = [np.zeros((20, 20, 20, 3)), np.zeros((20, 20, 19, 3))]

    indice = get_patch_indice_all(images, (10, 10, 10), 0.5)

    expected = []

    expected_indice_1 = [
        (0, 0, 0), (0, 0, 5), (0, 0, 10),
        (0, 5, 0), (0, 5, 5), (0, 5, 10),
        (0, 10, 0), (0, 10, 5), (0, 10, 10),
        (5, 0, 0), (5, 0, 5), (5, 0, 10),
        (5, 5, 0), (5, 5, 5), (5, 5, 10),
        (5, 10, 0), (5, 10, 5), (5, 10, 10),
        (10, 0, 0), (10, 0, 5), (10, 0, 10),
        (10, 5, 0), (10, 5, 5), (10, 5, 10),
        (10, 10, 0), (10, 10, 5), (10, 10, 10)
    ]

    expected_indice_2 = [
        (0, 0, 0), (0, 0, 5), (0, 0, 9),
        (0, 5, 0), (0, 5, 5), (0, 5, 9),
        (0, 10, 0), (0, 10, 5), (0, 10, 9),
        (5, 0, 0), (5, 0, 5), (5, 0, 9),
        (5, 5, 0), (5, 5, 5), (5, 5, 9),
        (5, 10, 0), (5, 10, 5), (5, 10, 9),
        (10, 0, 0), (10, 0, 5), (10, 0, 9),
        (10, 5, 0), (10, 5, 5), (10, 5, 9),
        (10, 10, 0), (10, 10, 5), (10, 10, 9)]

    for index in expected_indice_1:
        expected.append((0, index))

    for index in expected_indice_2:
        expected.append((1, index))

    for index, expected_index in zip(indice, expected):
        assert np.all(index == expected_index)


def test_check_drop_3d():
    image = np.zeros((10, 10, 10, 2))
    image[5:, :, :, 0] = 1
    image[5:, :5, :, 1] = 1

    images = np.array([image for _ in range(2)])

    indice = get_patch_indice((10, 10, 10), (5, 5, 5), 0)
    indice = list(product([0, 1], indice))

    res = check_drop(images, indice, (5, 5, 5), 0.4, None)
    expected = [
        False, False, False, False, True, True, True, True
    ] * 2

    assert np.all(res == expected)

    res = check_drop(images, indice, (5, 5, 5), 0.9, None)
    expected = [
        False, False, False, False, True, True, False, False
    ] * 2

    assert np.all(res == expected)

    res = check_drop(images, indice, (5, 5, 5), 0.9, 0)
    expected = [
        False, False, False, False, True, True, True, True
    ] * 2

    assert np.all(res == expected)


# def test_check_drop_3d_different_images():
#     image_1 = np.zeros((10, 10, 10, 2))
#     image_1[5:, :, :, 0] = 1
#     image_1[5:, :5, :, 1] = 1

#     image_2 = np.zeros((9, 10, 10, 2))
#     image_2[4:, :, :, 0] = 1
#     image_2[4:, :5, :, 1] = 1

#     images = np.array([image_1, image_2], dtype=object)

#     indice = get_patch_indice_all(images, (5, 5, 5), 0)

#     res = check_drop(images, indice, (5, 5, 5), 0.4, None)
#     print(res)
#     expected = [
#         False, False, False, False, True, True, True, True
#     ] * 2

#     assert np.all(res == expected)

#     res = check_drop(images, indice, (5, 5, 5), 0.9, None)
#     expected = [
#         False, False, False, False, True, True, False, False
#     ] * 2

#     assert np.all(res == expected)

#     res = check_drop(images, indice, (5, 5, 5), 0.9, 0)
#     expected = [
#         False, False, False, False, True, True, True, True
#     ] * 2

#     assert np.all(res == expected)


def test_check_drop_2d():
    image = np.zeros((10, 10, 2))
    image[5:, :, 0] = 1
    image[5:, :5, 1] = 1

    images = np.array([image for _ in range(2)])

    indice = get_patch_indice((10, 10), (5, 5), 0)
    indice = list(product([0, 1], indice))

    res = check_drop(images, indice, (5, 5), 0.4, None)

    expected = [
        False, False, True, True
    ] * 2

    assert np.all(res == expected)

    res = check_drop(images, indice, (5, 5), 0.9, None)
    expected = [
        False, False, True, False
    ] * 2

    assert np.all(res == expected)

    res = check_drop(images, indice, (5, 5), 0.9, 0)
    expected = [
        False, False, True, True
    ] * 2

    assert np.all(res == expected)


def test_stratified_index():
    values = np.array([0, 0, 0, 0, 0, 1, 1, 1])
    index = get_stratified_index(values, 4)

    assert len(index) == len(values)
    assert np.all(np.unique(index) == np.arange(len(values)))

    assert sum(values[index[:4]]) == 2
    assert sum(values[index[4:]]) == 1

    index = get_stratified_index(values, 3)

    assert len(index) == len(values)
    assert np.all(np.unique(index) == np.arange(len(values)))

    assert sum(values[index[:3]]) == 1
    assert sum(values[index[3:6]]) == 1
    assert sum(values[index[6:]]) == 1

    index = get_stratified_index(values, 2)

    assert len(index) == len(values)
    assert np.all(np.unique(index) == np.arange(len(values)))

    assert sum(values[index[:2]]) == 2
    assert sum(values[index[2:4]]) == 0
    assert sum(values[index[4:6]]) == 1
    assert sum(values[index[6:]]) == 0


def test_get_patches_3d():
    image = np.zeros((10, 10, 10, 2))
    image[5:, :, :, 0] = 1
    image[5:, :5, :, 1] = 1

    images = np.array([image for _ in range(2)])

    indice = get_patch_indice((10, 10, 10), (5, 5, 5), 0)

    # Only patches, in order
    patches = get_patches(images,
                          patch_indice=indice, patch_size=(5, 5, 5),
                          stratified=False, drop_fraction=0,
                          check_drop_channel=None)

    empty_patches = np.zeros((5, 5, 5, 2))
    half_patches = np.zeros((5, 5, 5, 2))
    half_patches[..., 0] = 1
    full_patches = np.ones((5, 5, 5, 2))

    expected = np.array([
        empty_patches, empty_patches, empty_patches, empty_patches,
        full_patches, full_patches, half_patches, half_patches
    ] * 2)

    assert len(patches) == 16
    assert np.all(patches == expected)

    # Only patches, in order, drop half
    patches = get_patches(images,
                          patch_indice=indice, patch_size=(5, 5, 5),
                          stratified=False, drop_fraction=0.4,
                          check_drop_channel=None)

    expected = np.array([
        full_patches, full_patches, half_patches, half_patches
    ] * 2)

    assert len(patches) == 8
    assert np.all(patches == expected)

    # Only patches, in order, only keep full
    patches = get_patches(images,
                          patch_indice=indice, patch_size=(5, 5, 5),
                          stratified=False, drop_fraction=0.9,
                          check_drop_channel=None)

    expected = np.array([
        full_patches, full_patches
    ] * 2)

    assert len(patches) == 4
    assert np.all(patches == expected)

    # Only patches, in order, drop half based on one channel
    patches = get_patches(images,
                          patch_indice=indice, patch_size=(5, 5, 5),
                          stratified=False, drop_fraction=0.4,
                          check_drop_channel=0)

    expected = np.array([
        full_patches, full_patches, half_patches, half_patches
    ] * 2)

    assert len(patches) == 8
    assert np.all(patches == expected)

    # patches, targets, stratified, without batch_size
    target = np.zeros((10, 10, 10, 1))
    target[5:, :5, :, 0] = 1

    targets = np.array([target for _ in range(2)])

    patches, labels = get_patches(images, target=targets,
                                  patch_indice=indice, patch_size=(5, 5, 5),
                                  stratified=True, drop_fraction=0.4,
                                  check_drop_channel=0)

    assert len(patches) == 8
    full_patches_num = 0
    half_patches_num = 0

    for patch, label in zip(patches, labels):
        if np.all(full_patches == patch):
            assert np.all(label == np.ones((5, 5, 5, 1)))
            full_patches_num += 1
        elif np.all(half_patches == patch):
            assert np.all(label == np.zeros((5, 5, 5, 1)))
            half_patches_num += 1

    assert full_patches_num == 4 and half_patches_num == 4

    # patches, targets, stratified based on target
    patches, labels = get_patches(images, target=targets,
                                  patch_indice=indice, patch_size=(5, 5, 5),
                                  stratified=True, drop_fraction=0.4,
                                  check_drop_channel=0, batch_size=4)

    assert len(patches) == 8

    full_patches_num = 0
    half_patches_num = 0
    for patch, label in zip(patches[:4], labels[:4]):
        if np.all(full_patches == patch):
            assert np.all(label == np.ones((5, 5, 5, 1)))
            full_patches_num += 1
        elif np.all(half_patches == patch):
            assert np.all(label == np.zeros((5, 5, 5, 1)))
            half_patches_num += 1

    assert full_patches_num == 2 and half_patches_num == 2

    full_patches_num = 0
    half_patches_num = 0
    for patch, label in zip(patches[4:], labels[4:]):
        if np.all(full_patches == patch):
            assert np.all(label == np.ones((5, 5, 5, 1)))
            full_patches_num += 1
        elif np.all(half_patches == patch):
            assert np.all(label == np.zeros((5, 5, 5, 1)))
            half_patches_num += 1

    assert full_patches_num == 2 and half_patches_num == 2


def test_get_patches_2d():
    image = np.zeros((10, 10, 2))
    image[5:, :, 0] = 1
    image[5:, :5, 1] = 1
    images = np.array([image for _ in range(2)])

    target = np.zeros((10, 10, 1))
    target[5:, :5, 0] = 1
    targets = np.array([target for _ in range(2)])

    indice = get_patch_indice((10, 10), (5, 5), 0)

    patches, labels = get_patches(images, target=targets,
                                  patch_indice=indice, patch_size=(5, 5),
                                  stratified=True, drop_fraction=0.4,
                                  check_drop_channel=0, batch_size=2)

    half_patches = np.zeros((5, 5, 2))
    half_patches[..., 0] = 1
    full_patches = np.ones((5, 5, 2))

    assert len(patches) == 4

    full_patches_num = 0
    half_patches_num = 0
    for patch, label in zip(patches[:2], labels[:2]):
        if np.all(full_patches == patch):
            assert np.all(label == np.ones((5, 5, 1)))
            full_patches_num += 1
        elif np.all(half_patches == patch):
            assert np.all(label == np.zeros((5, 5, 1)))
            half_patches_num += 1

    assert full_patches_num == 1 and half_patches_num == 1

    full_patches_num = 0
    half_patches_num = 0
    for patch, label in zip(patches[2:], labels[2:]):
        if np.all(full_patches == patch):
            assert np.all(label == np.ones((5, 5, 1)))
            full_patches_num += 1
        elif np.all(half_patches == patch):
            assert np.all(label == np.zeros((5, 5, 1)))
            half_patches_num += 1

    assert full_patches_num == 1 and half_patches_num == 1
